// sensible server which advertises itself via Bonjour

// NODE INCLUDES

var	dgram = require ("dgram");
var	fs = require ("fs");
var	http = require ("http");
var	os = require ("os");
var	url = require ("url");

// REGULAR JS INCLUDES

// assume that monohm.js lives in the same directory as our mainline
var	code = fs.readFileSync (require ("path").dirname (process.argv [1]) + "/monohm.js");
eval (code.toString ());

// assume that sensible.js lives in the same directory as our mainline
var	code = fs.readFileSync (require ("path").dirname (process.argv [1]) + "/sensible.js");
eval (code.toString ());

// MAINLINE

sensible.ApplicationFactory.createApplication
(
	function (inError)
	{
		if (inError)
		{
			console.error ("error during sensible application startup");
			console.error (inError);
		}
		else
		{
			console.log ("sensible application startup");
		}
	}
);

// called just before sensible.Application.start()
sensible.node.Application.prototype.onBeforeStart = function (inCallback)
{
	console.log ("node.Application.onBeforeStart()");

	// ok we need to find the Little Sky and set various things
	// before we allow Sensible to continue
	this.socket = dgram.createSocket ("udp4");

	// save away the callback
	// as we'll need it for when the discover command returns
	this.discoverCallback = inCallback;
	
	var	self = this;
	
	// note that later nodes require the address parameter
	this.socket.bind
	(
		0,
		"0.0.0.0",
		function (inError)
		{
			if (inError)
			{
				console.error ("cannot bind socket");
				console.error (inError);
			}
			else
			{
				// now broadcast our discover message
				self.socket.setBroadcast (true);
	
				var	message = "!discover;";
	
				// the documentation says send() takes a string or a buffer
				// it's lying
				var	messageBuffer = new Buffer (message);
	
				self.socket.send
				(
					messageBuffer,
					0,
					messageBuffer.length,
					50007,
					"255.255.255.255",
					function (inError)
					{
						if (inError)
						{
							console.error (inError);
						}
					}
				);
				
				self.socket.on
				(
					"message",
					function (inMessage, inRemoteInfo)
					{
						var	messageString = inMessage.toString ();
						
						if (messageString.charAt (0) == "!")
						{
							console.log ("little sky message received");
							console.log (messageString);
							
							// strip the opening ! and trailing ;
							messageString = messageString.slice (1, messageString.length - 1);
							
							var	messageElements = messageString.split (",");
							var	messageStatus = messageElements.splice (0, 1) [0];
							
							if (messageStatus == "OK")
							{
								var	messageOpcode = messageElements.splice (0, 1) [0];
								messageOpcode = messageOpcode.toLowerCase ();
								
								// find a handler!
								var	handlerName = "ls_message_" + messageOpcode;
								
								if (typeof (self [handlerName]) == "function")
								{
									self [handlerName].call (self, messageElements);
								}
								else
								{
									console.error ("couldn't find handler " + handlerName);
								}
							}
							else
							{
								console.error ("message status error");
							}
						}
						else
						{
							console.error ("non-LS message received");
							console.error (messageString);
						}
					}
				);
			}
		}
	);
}

// called just after sensible.Application.start()
sensible.node.Application.prototype.onAfterStart = function (inCallback)
{
	console.log ("node.Application.onAfterStart()");
	
	inCallback ();
}

// LS MESSAGE HANDLERS

sensible.node.Application.prototype.ls_message_discover = function (inParameters)
{
	console.log ("ls_message_discover()");
	
	this.littleSkyAddress = inParameters [0];
	this.littleSkyNickname = inParameters [1];
	
	// copy the nickname to the config
	// so that's what shows up in MDNS searches
	this.config.name = this.littleSkyNickname;
	
	console.log ("found LS name of " + this.config.name);
	
	// if we have a callback cached, call it to allow app startup
	if (this.discoverCallback)
	{
		console.log ("calling discover callback");
		this.discoverCallback.call (this);
		this.discoverCallback = null;
	}
}

// OVERRIDES

// we override all property set calls
sensible.node.Application.prototype.properties_set = function (inRequest, inCallback)
{
	console.log ("Application.properties_set()");
	
	for (var param in inRequest.parameters)
	{
		if (typeof param == "string")
		{
			if (this.propertiesByKey [param])
			{
				var	value = inRequest.parameters [param];

				if (typeof (value) == "string")
				{
					if (param == "brightness")
					{
						var	brightness = parseInt (value);
						console.log ("setting brightness to " + brightness);
					
						this.sendPacket ("!setbr," + brightness + ";");
					}
					else
					if (param == "br_up")
					{
						this.sendPacket ("!br_up;");
					}
					else
					if (param == "br_down")
					{
						this.sendPacket ("!br_down;");
					}
					else
					if (param == "next_playlist")
					{
						this.sendPacket ("!next_playlist;");
					}
					else
					{
						console.log ("no handler for registered param: " + param);
					}
				}
			}
		}
		
		console.log (param + " = " + inRequest.parameters [param]);
	}
	
	var	response = new Object ();
	response.type = "json";
	response.object = null;
	
	inCallback (response);
}

// LITTLESKY APIS

sensible.node.Application.prototype.sendPacket = function (inMessageText)
{
	console.log ("sendPacket(" + inMessageText + ")");
	
	var	messageBuffer = new Buffer (inMessageText);
	
	this.socket.send
	(
		messageBuffer,
		0,
		messageBuffer.length,
		50007,
		this.littleSkyAddress,
		function (inError)
		{
			if (inError)
			{
				console.error (inError);
			}
		}
	);
}

